/*
  camlib.c
    CAMAC Library for K3922 & K2917
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include <signal.h>

#define SIGVME SIGUSR1   
/* use SIGUSR1 for VME Interrupt Signal */

#include <vmehb_ioctl.h> /* vmehb driver ioctl */

#include "k2917.h"
#include "camlib.h"

#define VME_A16D16BASE 0x0000000
#define VME_A16D16SIZE 0xFFFF

#define NTRY 100

#define K3922SR_WMASK 0x000387

static int fdcam=-1; 
static void *cam=0;
static int VMEoffset=0;
static int VMElength=0;
static void  *VMEaddress=0; 

static unsigned short branch=0,crate=0;


inline void _kin_write( int offset, unsigned short data )
{
  unsigned short *addr;

  addr=(unsigned short *)(cam+offset);
#if defined(DEBUG) || defined(DEBUGLIST)  
  fprintf(stderr,"write %8x <--- %4hx\n",(int)addr,data);
#endif
  *addr=data;
}

inline unsigned short _kin_read( int offset )
{
  unsigned short *addr,data;

  addr=(unsigned short *)(cam+offset);
  data=(*addr);
#if defined(DEBUG) || defined(DEBUGLIST) 
  fprintf(stderr,"read  %8x ---> %4hx\n",(int)addr,data);
#endif
  return data;
}


#if defined(DEBUG) || defined(DEBUGINT) || defined(DEBUGLIST)
inline unsigned short _kin_read2( int offset )
{
  unsigned short *addr;
  addr=(unsigned short *)(cam+offset);
  return *addr;
}

int printIntRegister( void )
{
  fprintf(stderr,
	  "ICR --> LAM: %4hx DONE: %4hx EMPTY: %4hx ABORT: %4hx\n",
	  _kin_read2(LAMICR),_kin_read2(DONEICR),
	  _kin_read2(EMPTYICR),_kin_read2(ABTICR));
  fprintf(stderr,
	  "IVR --> LAM: %4hx DONE: %4hx EMPTY: %4hx ABORT: %4hx\n",
	  _kin_read2(LAMIVR),_kin_read2(DONEIVR),
	  _kin_read2(EMPTYIVR),_kin_read2(ABTIVR));
  return 0;
}
#endif

int COPEN( void )
{
  static const char *funcname="COPEN";
  int pagesize;

  pagesize=getpagesize();
  if(fdcam!=-1){
    fprintf(stderr,"[%s]: CAMAC already opened. Try to repen \n",
	    funcname);
    if(VMEaddress)
      munmap((void *)VMEaddress,(size_t)VMElength);
    close(fdcam);
  }
  /* open VME interface */
  if((fdcam=open("/dev/vme16d16",O_RDWR))==-1){
    fprintf(stderr,"[%s]: VME open failure\n",funcname);
    return 1;
  }
  /* mmap for K2917 */
  if(K2917SIZE>VME_A16D16SIZE || K2917ADR<VME_A16D16BASE
     || (K2917ADR+K2917SIZE)>(VME_A16D16BASE+VME_A16D16SIZE)){
    fprintf(stderr,"[%s]: Invalid Address Base:%8x Size:%8x\n",
	    funcname,K2917ADR,K2917SIZE);
    close(fdcam); fdcam=-1;
    return 1;
  }
  VMEoffset=K2917ADR-VME_A16D16BASE;
  VMElength=K2917SIZE+VMEoffset%pagesize;
#ifdef DEBUG
  fprintf(stderr,"[%s]: PageSize:%8x Offset:%8x MapSize:%8x\n",
	  funcname,pagesize,VMEoffset,VMElength);
#endif
  VMEaddress=mmap( 0, (size_t)VMElength, PROT_READ | PROT_WRITE, 
		   MAP_SHARED,fdcam,VMEoffset/pagesize*pagesize);
#ifdef DEBUG
  fprintf(stderr,"[%s]: mmap call RetValue=%8x\n",
	 funcname,(int)VMEaddress);
#endif

  if(VMEaddress==0){
    fprintf(stderr,"[%s]: mmap fails\n",funcname);
    close(fdcam); fdcam=-1;
    return 1;
  }
  cam=VMEaddress+(VMEoffset%pagesize);

#ifdef DEBUG
  fprintf(stderr,"[%s]: cam=%8x\n",funcname,(int)cam);

  fprintf(stderr,"before RESET ...\n"); 
  printIntRegister();
#endif


  _kin_write(CSR,0x1000);
#ifdef DEBUG
  fprintf(stderr,"after RESET ...\n"); 
  printIntRegister();
#endif
  _kin_write(CSER,0);
  _kin_write(DOCR,0x8010);
  _kin_write(SCCR,0);
  _kin_write(MTCR,0);

  _kin_write(LAMICR,0);
  _kin_write(DONEICR,0);
  _kin_write(EMPTYICR,0);
  _kin_write(ABTICR,0);

  _kin_write(LAMIVR,0x0F);
  _kin_write(DONEIVR,0x0F);
  _kin_write(EMPTYIVR,0x0F);
  _kin_write(ABTIVR,0x0F);

  _kin_write(AMR,0x79);
  _kin_write(CMAR,0x1FFF);
  _kin_write(CMR,0x80);        /* Halt Command */
      
  return 0;
}

int CCLOSE( void )
{
  if(VMEaddress){
    munmap( VMEaddress, (size_t)VMElength );
    VMEaddress=0; VMEoffset=0; VMElength=0;
  }
  close(fdcam); fdcam=-1;
  return 0;
}

int CRESET( void )
{
  _kin_write(CSR,0x1000);
  return 0;
}

int CSETBR( int br )
{
  if((br&0x7)!=br)
    return -1;
  branch=br;
  return 0;
}

int CSETCR( int cr )
{
  if((cr&0x7)!=cr)
    return -1;
  crate=cr;
  return 0;
}

int CGENZ( void )
{
  int status,q,x;
  short data=0;

  if((status=CAMACW(NAF(30,0,1),&data,&q,&x))!=0)
     return status;
  data=(data|0x1)&K3922SR_WMASK;
  return(CAMACW(NAF(30,0,17),&data,&q,&x));
}

int CGENC( void )
{
  int status,q,x;
  short data=0;

  if((status=CAMACW(NAF(30,0,1),&data,&q,&x))!=0)
     return status;
  data=(data|0x2)&K3922SR_WMASK;
  return(CAMACW(NAF(30,0,17),&data,&q,&x));
}

int CSETI( void )
{
  int status,q,x;
  short data=0;
  if((status=CAMACW(NAF(30,0,1),&data,&q,&x))!=0)
    return status;
  data=(data|0x4)&K3922SR_WMASK;
  return(CAMACW(NAF(30,0,17),&data,&q,&x));
}

int CREMI( void )
{
  int status,q,x;
  short data=0;
  if((status=CAMACW(NAF(30,0,1),&data,&q,&x))!=0)
    return status;
  data=data&(~0x4)&K3922SR_WMASK;
  return(CAMACW(NAF(30,0,17),&data,&q,&x));
}



static int _kin_sst_ptro( int naf, int ns, unsigned short *dlow,
			  unsigned short *dhigh, int *q, int *x )
{
  volatile unsigned short status;
  int i;

  /* reset the error bit */
  //  _kin_write(CSR,0x1000);
  /*
  _kin_write(CSER,0);
  */
  /* set CMAR to beginning */ 
  _kin_write(CMAR,0x1FFD);
  /* load data in command memory */
  _kin_write(CMR,(crate<<8)|((ns&0x1)<<1)|0x1);
  _kin_write(CMR,naf);
  _kin_write(CMR,0x80);
  /* set CMAR to beginning */ 
  _kin_write(CMAR,0x1FFD);
  _kin_write(CSR,0x01);

  (*q)=(*x)=0;
  do{
    i=-1;
    do{
      i++; status=_kin_read(CSR);
#ifdef DEBUG
      fprintf(stderr,"[PTRO1]: %d %4hx\n",i,status);
#endif
    }while((status&0x8180)==0 && i<NTRY);

    if(i>=NTRY) return ETIMEDOUT;

    if((status&0x0100)){
      *dlow=_kin_read(DLR);
      if(ns==0) *dhigh=_kin_read(DHR);
      status=_kin_read(CSR);
    }
  }while((status&0x8080)==0);

  *q=(status&0x2)?0:1;
  *x=(status&0x4)?0:1;

  if(status&0x2000) return ETIMEDOUT;

  return 0;
}

static int _kin_sst_ptwo( int naf, int ns, unsigned short *dlow,
			  unsigned short *dhigh, int *q, int *x )
{
  volatile unsigned short status;
  int i;

  /* reset the error bit */
  //  _kin_write(CSR,0x1000); 
  //  _kin_write(CSER,0);
  /* set CMAR to beginning */ 
  _kin_write(CMAR,0x1FFD);
  /* load data in command memory */
  _kin_write(CMR,(crate<<8)|((ns&0x1)<<1)|0x1);
  _kin_write(CMR,naf);
  _kin_write(CMR,0x80);
  /* set CMAR to beginning */ 
  _kin_write(CMAR,0x1FFD);
  _kin_write(CSR,0x21);

  (*q)=(*x)=0;
  do{
    i=-1;
    do{
      i++; status=_kin_read(CSR);
#ifdef DEBUG
      fprintf(stderr,"[PTWO1]: %d %4hx\n",i,status);
#endif
    }while((status&0x8180)==0 && i<NTRY);

    if(i>=NTRY) return ETIMEDOUT;

    if((status&0x0100)){
      _kin_write(DLR,*dlow);
      if(ns==0) _kin_write(DHR,*dhigh);
      status=_kin_read(CSR);
    }
  }while((status&0x8080)==0);

  *q=(status&0x2)?0:1;
  *x=(status&0x4)?0:1;

  if(status&0x2000) return ETIMEDOUT;
  return 0;
}

static int _kin_sst_ptco( int naf, int *q, int *x )
{
  volatile unsigned short status;
  int i;

  /* reset the error bit */
  //  _kin_write(CSR,0x1000); 
  //  _kin_write(CSER,0);
  /* set CMAR to beginning */ 
  _kin_write(CMAR,0x1FFD);
  /* load data in command memory */
  _kin_write(CMR,(crate<<8)|0x1);
  _kin_write(CMR,naf);
  _kin_write(CMR,0x80);
  /* set CMAR to beginning */ 
  _kin_write(CMAR,0x1FFD);
  _kin_write(CSR,0x01);

  (*q)=(*x)=0;
  i=-1;
  do{
    i++; status=_kin_read(CSR);
#ifdef DEBUG
    fprintf(stderr,"[PTCO1]: %d %4hx\n",i,status);
#endif
  }while((status&0x8080)==0 && i<NTRY );

  if(i>=NTRY) return ETIMEDOUT;

  *q=(status&0x2)?0:1;
  *x=(status&0x4)?0:1;

  if(status&0x2000) return ETIMEDOUT;
  return 0;
}
      
int CAMACL( int naf, int *data, int *q, int *x )
{
  unsigned short low,high;
  int status;
#ifdef DEBUG
  static const char *funcname="CAMACL"; 
  int n,a,f;

  n=GetN(naf); a=GetA(naf); f=GetF(naf);
  fprintf(stderr,"[%s]: N:%2d A=%2d F=%2d Data:%8x\n",
	  funcname,n,a,f,*data);
#endif
  
  if(CHKF8(naf)){       /* Control Operation */
    status=_kin_sst_ptco(naf,q,x);
  }
  else if(CHKF16(naf)){ /* Write Operation */
    low=(*data)&0xFFFF; high=((*data)>>16)&0xFF;
    status=_kin_sst_ptwo(naf,0,&low,&high,q,x);
  }
  else{                 /* Read Operation */
    status=_kin_sst_ptro(naf,0,&low,&high,q,x);
    *data=((high&0xFF)<<16)|(low&0xFFFF);
  }
#ifdef DEBUG
    fprintf(stderr,"[%s]: St:%8x Q:%d X:%d Data:%8x \n",
	    funcname,status,*q,*x,*data);
#endif
  return status;
}

int CAMACW( int naf, short *data, int *q, int *x )
{
  unsigned short low,high=0;
  int status;
#ifdef DEBUG
  static const char *funcname="CAMACW"; 
  int n,a,f;

  n=GetN(naf); a=GetA(naf); f=GetF(naf);
  fprintf(stderr,"[%s]: N:%2d A=%2d F=%2d Data:%4hx\n",
	  funcname,n,a,f,*data);
#endif
  
  if(CHKF8(naf)){       /* Control Operation */
    status=_kin_sst_ptco(naf,q,x);
  }
  else if(CHKF16(naf)){ /* Write Operation */
    low=(unsigned short)((*data)&0xFFFF);
    status=_kin_sst_ptwo(naf,1,&low,&high,q,x);
  }
  else{                 /* Read Operation */
    status=_kin_sst_ptro(naf,1,&low,&high,q,x);
    *data=(short)((int)low&0xFFFF);
  }
#ifdef DEBUG
    fprintf(stderr,"[%s]: St:%8x Q:%d X:%d Data:%4hx \n",
	    funcname,status,*q,*x,*data);
#endif
  return status;
}

static volatile int int_counter=0;

static void sig_handler()
{
  int_counter=1;
  signal(SIGVME,sig_handler);
#ifdef DEBUGINT
  fprintf(stderr,"[handler]: got signal\n");
#endif
}


int CENLAM( int mask )
{
  int q,x,i;
#ifndef NOVMEI
  struct irq_data irqd;
  int status;
#endif

  int_counter=0;
#ifdef NOVMEI
  return(CAMACL(NAF(30,13,17),&mask,&q,&x));
#else
  signal(SIGVME,sig_handler);

  irqd.level=INTR_LEVEL;
  irqd.vector=INTR_VECTOR;
  irqd.sig=SIGVME;
  irqd.response=VME_NACK;
  irqd.handler=NULL;
  status=ioctl(fdcam,VME_IRQ_REQUEST,&irqd);
#ifdef DEBUGINT
  fprintf(stderr,"[CENLAM]: IRQ -> Level:%d Vector:%d Signal:%d \n",
	  irqd.level,irqd.vector,irqd.sig);
  fprintf(stderr," RetVal of ioctl(VME_IRQ_REQUEST) : %d\n",status); 
#endif

  _kin_write(LAMIVR,INTR_VECTOR);
  _kin_write(LAMICR,0x18+INTR_LEVEL);
  CAMACL(NAF(30,13,17),&mask,&q,&x);
  CAMACL(NAF(30,0,1),&i,&q,&x);
  i=(i|0x100)&K3922SR_WMASK;
  status=CAMACL(NAF(30,0,17),&i,&q,&x);
#ifdef DEBUGINT
  fprintf(stderr,"[CENLAM]: LAM Mask:%8x St:%8x\n",mask,status);
#endif
  return(status);
#endif
}

int CDSLAM( void )
{
  int q,x,i,st;

#ifndef NOVMEI
  struct irq_data irqd;
#endif

  i=0;
  CAMACL(NAF(30,13,17),&i,&q,&x);
  CAMACL(NAF(30,0,1),&i,&q,&x);
  i=(i&K3922SR_WMASK&(~0x100));
  st=CAMACL(NAF(30,0,17),&i,&q,&x);
#ifndef NOVMEI
  _kin_write(LAMICR,0);
  _kin_write(LAMIVR,0x0F);
  irqd.level=INTR_LEVEL;
  irqd.vector=INTR_VECTOR;
  irqd.sig=SIGVME;
  irqd.handler=NULL;
  irqd.response=VME_NACK;
  st=ioctl(fdcam,VME_IRQ_FREE,&irqd);
#ifdef DEBUG
  fprintf(stderr,"RetVal of ioctl(VME_IRQ_FREE) : %d\n",st); 
#endif

  signal(SIGVME,SIG_DFL);
#endif
  int_counter=0;
  return st;
}

int CWTLAM( int timeout )
{
  unsigned short a;
  unsigned int rtime;

#ifdef NOVMEI

#else
  /* reset the error bit */
  //  _kin_write(CSR,0x1000);
  /* enable interrupt */
  a=_kin_read(LAMICR);
  _kin_write(LAMICR,(a|0x10));
  if(int_counter==0){
    rtime=sleep((unsigned int)timeout);
    if(rtime==0){ /* Timeout */
#ifdef DEBUGINT
      fprintf(stderr,"[CWTLAM]: Timeout \n");
#endif
      return ETIMEDOUT;
    }
  }
  if(int_counter==1){
    int_counter=0;
    a=_kin_read(CSR);
#ifdef DEBUGINT
    fprintf(stderr,"[CWTLAM]: Got Signal. CSR=%4hx\n",a);
#endif
    if(a&0x200){   /* LAM Interrupt */
      return 0;
    }
    else{          /* Unknown Errors */
      return EIO;
    }
  }
  else{                /* Get Other Signals */ 
    return EINTR;
  }
#endif

}

#define MAXLIST 10
struct ListSt{
  int size;
  int *list;
};
static struct ListSt clist[MAXLIST];

int CAMRegisterList( int id, int size, int bufsize, int *alist )
{
  static const char *funcname="CAMRegisterList";
  int *lp;
  int i;

  if(id<0 || id>=MAXLIST){
    fprintf(stderr,"[%s]: Invalid List ID %d \n",funcname,id);
    return EINVAL;
  }
  if(clist[id].list){
    fprintf(stderr,"[%s]: List ID %d is already registered\n",
	    funcname,id);
    return EINVAL;
  }
  if((lp=malloc(sizeof(int)*size))==0){
    fprintf(stderr,"[%s]: malloc fails\n",funcname);
    return ENOBUFS;
  }
  clist[id].list=lp; clist[id].size=size;
  for(i=0;i<size;i++)
    lp[i]=alist[i];

  return 0;
}

int CAMUnregisterList( int id )
{
  static const char *funcname="CAMUnregisterList";


  if(id>=0 && id<MAXLIST && clist[id].list){
    free((void *)clist[id].list);
    clist[id].list=0; clist[id].size=0;
  }
  else{
    fprintf(stderr,"[%s]: Invalid List ID %d\n",funcname,id);
    return EINVAL;
  }
  return 0;
}

int CAMBlockRead( int id, int maxbufsize, int *buf )
{
  static const char *funcname="CAMBlockRead";
  int i,w,n,a,j;
  int naf,qs;
  unsigned short low;
  int data,q,x;
  volatile unsigned short status;

  if(id<0 || id>=MAXLIST || clist[id].list==0){
    fprintf(stderr,"[%s]: Invalid List ID %d\n",funcname,id);
    return 0;
  }
  w=0;
  for(i=0;i<clist[id].size;i++){
    naf=clist[id].list[i]&NAFREAD; qs=clist[id].list[i]&QS_READ;
    n=GetN(clist[id].list[i]); a=GetA(clist[id].list[i]);
    if(qs){  /* Q-Stop Read */
      /* reset the error bit */
      //      _kin_write(CSR,0x1000);
      /* set CMAR to beginning */
      _kin_write(CMAR,0x1FFB);
      /* load data in command memory */
      _kin_write(CMR,(crate<<8)|0x22);  /* Q-Stop Block Transfer */
      _kin_write(CMR,naf);
      _kin_write(CMR,0x07FF);     /* set max # data to 2K */
      _kin_write(CMR,0xFFFF);
      _kin_write(CMR,0x80);
      /* set CMAR to beginning */
      _kin_write(CMAR,0x1FFB);
      _kin_write(CSR,0x01);
      do{
	j=-1;
	do{
	  status=_kin_read(CSR); j++;
	}while((status&0x8180)==0&&j<NTRY);
	if(j>=NTRY) status|=0x2000;
	if((status&0x0100)){
	  low=_kin_read(DLR);
	  status=_kin_read(CSR);
	  q=(status&0x2)?0:1;
	  if(q){
	    x=(status&0x4)?0:1;
	    data=LDencode(q,x,n,a,low);
	    if(w<maxbufsize) buf[w++]=data;
	    else             return w;
	  }
	}
      }while(status&0xE080);
    }
    else{    /* Normal Read */
      /* reset the error bit */
      //      _kin_write(CSR,0x1000);
      //_kin_write(CSER,0);
      /* set CMAR to beginning */
      _kin_write(CMAR,0x1FFD);
      /* load data in command memory */
      _kin_write(CMR,(crate<<8)|0x2);   /* 16 bit transfer */
      _kin_write(CMR,naf);
      _kin_write(CMR,0x80);
      /* set CMAR to beginning */
      _kin_write(CMAR,0x1FFD);
      _kin_write(CSR,0x01);
      j=-1;
      do{
	status=_kin_read(CSR); j++;
      }while((status&0x8180)==0 && j<NTRY);
      if((status&0x0100)){
	low=_kin_read(DLR);
	status=_kin_read(CSR);
	q=(status&0x2)?0:1;
	if(q){
	  x=(status&0x4)?0:1;
	  data=LDencode(q,x,n,a,low);
	  if(w<maxbufsize)  buf[w++]=data;
	  else    	    return w;
	}
      }
    }
  }  
  return w;
}
